"""
Copyright (C) 2019 Cumulus Networks, Inc. all rights reserved

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
Cumulus Linux systems: /usr/share/cumulus/EULA.txt
"""

import re
import logging
from pprint import pformat

"""
parser for topology.dot
"""

this_log = logging.getLogger(__name__)
debug    = this_log.debug
info     = this_log.info
warn     = this_log.warning
error    = this_log.error

name = "cat topology.dot"
aliases = ['topology.dot']


def parse(command, output):
    """
    Return a dictionary or list containing the values parsed from the output
    If parse failed, return None

    Note that this parser uses line breaks to delimit data, which is not a requirement for the
    dot format.  A fully featured dot format parser would be better, but would involve
    either including new libraries like pydot and pyparsing into CL, or calling into the
    C library libcgraph6 which is already in the distribution.
    """

    '''
  graph G {
    # Headers (ID type)
    LLDP=""
    BFD="upMinTx=150,requiredMinRx=250,afi=both"
    
    # Hosts (Node type)
    "oob-mgmt-switch"  [function="oob-switch" mgmt_ip="192.168.200.2" version="3.7.9" ssh_user="cumulus" tunnelip="172.31.255.254"]
    "oob-mgmt-server"  [function="oob-server" mgmt_ip="192.168.200.1" ssh_user="cumulus" tunnelip="172.31.255.254"]
    "netq-ts"   [function="host" mgmt_ip="192.168.200.250" tunnelip="172.31.255.254" cpu="4" memory="8192" os="cumulus/tscloud222" ssh_user="cumul
us" ztp="../tc_files/netq-cloud-ztp.sh" config="./helper_scripts/extra_switch_config.sh"]

    "BDR0001" [function="tor" version="3.7.9" ssh_user="cumulus" tunnelip="172.31.255.254"]
    "BDR0002" [function="tor" version="3.7.9" ssh_user="cumulus" tunnelip="172.31.255.254"]

    # Connections/Links (Edge Type)
    "dell-s4000-10":"swp38" -- "qct-ly4r-04":"swp23" [LLDP="match_type=portdescr"]
    "dell-s4000-10":"swp49" -- "mlx-2700b-03":"swp5" [LLDP="match_type=portdescr"]
    "cumulus":"swp44" -- "qct-ly2-04":"swp20" [BFD="bfdtmpl=BFD1", LLDP="lldptmpl=LLDP1"]
    "dell-s6000-22":"swp20s0" -- "layer1-automation":"p72p2" [LLDP="match_type=portdescr"]
    "S2-C1.MSN2700.domain":"swp1" -- "S1-R1.MSN2410.domain":"swp49" [LLDP="match_hostname=fqdn"];
  }`3
    '''
    parsed_dct = {}
    hdr_pattern = re.compile(r'''^\s*(strict)?\s*((di)?(sub)?graph.*|\s*\}|\s*\])\s*;?\s*$''', re.IGNORECASE)
    blank = re.compile(r'''^\s*$''')
    comment = re.compile(r'''^\s*(\#|//|/\*|\*/).*$''')
    global_param_pattern = re.compile(r'''^
            \s*{?
            \s*(?P<keyword>("[a-zA-Z0-9 _:\-+./()]+"|[a-zA-Z0-9_]+))
            \s*=
            \s*(?P<value>("[a-zA-Z0-9 _:=,;\-\#+./()]*"|[a-zA-Z0-9_\-.]+))
            \s*(;|,)?
            \s*$''', re.X)

    node_stmt_pattern = re.compile(r'''^
            \s*{?
            \s*(?P<node>("[a-zA-Z0-9 _:\-+./()]+"|[a-zA-Z0-9_]+))
            (\s*(?P<node_attributes>\[[\-._/ :,=";\#\w]+\]))?
            ;?
            \s*$''', re.X)
    # Hack to catch attribute statement lists that start with node [, edge [
    attribute_header_pattern = re.compile(r'''^
            \s*(?P<type>(node|edge))\s*(?P<attributes>\[.*)?$
            ''', re.X | re.IGNORECASE)

    link_pattern = re.compile(r'''^
            \s*{?
            \s*(?P<dut> ("[a-zA-Z0-9 _:\-+./()]+"|[a-zA-Z0-9_\-.]+))\s*:
            \s*(?P<port>("[a-zA-Z0-9 _:\-+./()]*"|[a-zA-Z0-9_\-.]+))
            \s*-[->]
            \s*(?P<neighbor_dut> ("[a-zA-Z0-9 _:\-+./()]+"|[a-zA-Z0-9_\-.]+))
            (\s*:\s*(?P<neighbor_port>("[a-zA-Z0-9 _:\-+./()]*"|[a-zA-Z0-9_\-.]+)))?
            \s*(?P<port_param>\[[\-._/ :,=";\w]+\])?
            \s*;?
            (\s*(?P<comment_at_end>(/\*|\#|//).*))?
            \s*$''', re.X)
    # TODO create a list of all the params used on the link or global, and store with a key
    # For now, it will just be a string
    # port_param_pattern = re.compile(r'''(?P<keyword>\w+)="?(?P<value>)"?.*\])''')

    for line in output.split('\n'):
        # ignore the header and trailer:
        if hdr_pattern.match(line):
            debug('matched hdr')
            continue
        if blank.match(line):
            debug('blank line')
            continue
        if comment.match(line):
            debug('comment line')
            continue

        # The primary goal of this parser is to process the links, we don't care about the other attributes
        # Because the parser currently works line by line vs delimited segments, some of the non-link attributes
        # may end up in the wrong groups
        global_param_match = global_param_pattern.match(line)
        if global_param_match:
            debug('Matched global_param')
            if 'global_params' not in parsed_dct:
                parsed_dct['global_params'] = {}
            parsed_dct['global_params'][global_param_match.group('keyword')] = global_param_match.group('value')
            continue

        node_stmt_match = node_stmt_pattern.match(line)
        if node_stmt_match:
            debug('matched node statement')
            if 'node_statements' not in parsed_dct:
                parsed_dct['node_statements'] = {}
            parsed_dct['node_statements'][node_stmt_match.group('node')] = node_stmt_match.group('node_attributes')
            continue

        # Note, this is currently a hack, it will not collect all attributes in the list under this leaf
        # at this time, no dependencies on this lib care about anything but links
        attrib_stmt_match = attribute_header_pattern.match(line)
        if attrib_stmt_match:
            debug('matched attribute statement')
            if 'attr_statements' not in parsed_dct:
                parsed_dct['attr_statements'] = {}
            parsed_dct['attr_statements'][attrib_stmt_match.group('type')] = attrib_stmt_match.group('attributes')
            continue

        # Done: What dict format do I want to return?
        # topology_dct[dut][port] = (neighbor_dut, neighbor_port)

        link_match = link_pattern.match(line)
        if link_match:
            debug('Matched link (edge) pattern')
            debug('Match: ' + pformat(link_match.group(0)))
            if 'links' not in parsed_dct:
                parsed_dct['links'] = {}
            dut = link_match.group('dut')
            port = link_match.group('port')
            if dut not in parsed_dct['links']:
                parsed_dct['links'][dut] = {}
            parsed_dct['links'][dut][port] = dict(neighbor_dut=link_match.group('neighbor_dut'),
                                                  neighbor_port=link_match.group('neighbor_port'),
                                                  port_param=link_match.group('port_param'))

        else:
            # make sure to error and return 'None' if any line doesn't match
            error('No {} parser match: {}'.format(name, line))
            return None

    return parsed_dct
